package quimufu.colourful_portals;

import net.minecraft.block.Block;
import net.minecraft.block.BlockState;
import net.minecraft.block.FluidFillable;
import net.minecraft.block.ShapeContext;
import net.minecraft.entity.Entity;
import net.minecraft.entity.player.PlayerEntity;
import net.minecraft.fluid.Fluid;
import net.minecraft.fluid.FluidState;
import net.minecraft.item.ItemPlacementContext;
import net.minecraft.item.Items;
import net.minecraft.server.world.ServerWorld;
import net.minecraft.state.StateManager;
import net.minecraft.state.property.EnumProperty;
import net.minecraft.state.property.Properties;
import net.minecraft.util.BlockRotation;
import net.minecraft.util.DyeColor;
import net.minecraft.util.math.*;
import net.minecraft.util.shape.VoxelShape;
import net.minecraft.util.shape.VoxelShapes;
import net.minecraft.world.BlockView;
import net.minecraft.world.World;
import net.minecraft.world.WorldAccess;
import org.jetbrains.annotations.Nullable;
import quimufu.colourful_portals.util.CollisionAwareShapeContext;

import static quimufu.colourful_portals.ColourfulPortalsMod.PORTAL_MANAGER;


public class PortalBlock extends Block implements FluidFillable {
    public static final EnumProperty<Direction.Axis> AXIS = Properties.AXIS;
    public static final EnumProperty<DyeColor> DYE_COLOR = EnumProperty.of("colour", DyeColor.class);

    public static final VoxelShape X_AABB = Block.createCuboidShape(
            7.0D,
            0.0D,
            0.0D,
            9.0D,
            16.0D,
            16.0D
    );
    public static final VoxelShape Y_AABB = Block.createCuboidShape(
            0.0D,
            7.0D,
            0.0D,
            16.0D,
            9.0D,
            16.0D
    );
    public static final VoxelShape Z_AABB = Block.createCuboidShape(
            0.0D,
            0.0D,
            7.0D,
            16.0D,
            16.0D,
            9.0D
    );

    public PortalBlock(Settings settings) {
        super(settings);
        this.setDefaultState(this.stateManager.getDefaultState().with(AXIS, Direction.Axis.X).with(DYE_COLOR, DyeColor.BLACK));
    }

    @Override
    public VoxelShape getOutlineShape(BlockState state, BlockView world, BlockPos blockPos, ShapeContext shapeContext) {
        if (shapeContext.isHolding(Items.DEBUG_STICK)
            || shapeContext.isHolding(ColourfulPortalsMod.PORTAL_BLOCK_ITEM)
            || shapeContext.isHolding(ColourfulPortalsMod.PORTAL_FLUID_BUCKET_ITEM)) {
            return switch (state.get(AXIS)) {
                case Z -> Z_AABB;
                case Y -> Y_AABB;
                default -> X_AABB;
            };
        }
        return VoxelShapes.empty();
    }

    protected void onEntityCollisionOld(BlockState state, World world, BlockPos pos, Entity entity) {
        ColourfulPortalsMod.LOGGER.info("detected collision {}", entity.getType());
        if (entity.getBoundingBox()
                .intersects(getShape(state, world, pos).getBoundingBox())) {
            Vec3d vec3d = new Vec3d(0.5, 0.5f, 0.5);
            entity.slowMovement(state, vec3d);
        }
        Vec3d entityPos = entity.getPos();
        Vec3d prevEntityPos = new Vec3d(entity.prevX, entity.prevY, entity.prevZ);
        if(entityPos.equals(prevEntityPos) && entity.getVelocity().squaredDistanceTo(Vec3d.ZERO) >= 0.01D){
            //in this case, the collision logic was called prior to the movement logic,
            // so we'll have to calculate our own next postion
            entityPos = entityPos.add(entity.getVelocity());
        }
        Vec3d movement = entityPos.subtract(prevEntityPos);
        if (world instanceof ServerWorld) {
            Direction.Axis axis = state.get(AXIS);
            double centerPosAlongAxis = pos.toCenterPos().getComponentAlongAxis(axis);
            double entityPosAlongAxis = entityPos.getComponentAlongAxis(axis);
            double prevEntityPosAlongAxis = prevEntityPos.getComponentAlongAxis(axis);
            double movementAlongAxis = entityPosAlongAxis - prevEntityPosAlongAxis;
            if (prevEntityPosAlongAxis > entityPosAlongAxis) {
                if (entityPosAlongAxis < centerPosAlongAxis
                    && prevEntityPosAlongAxis >= centerPosAlongAxis) {
                    double plainHitAfter = (centerPosAlongAxis - entityPosAlongAxis) / movementAlongAxis;
                    Vec3d planeHitPoint = entityPos.add(movement.multiply(plainHitAfter));
                    if (Box.from(new BlockBox(pos)).contains(planeHitPoint)) {
                        PORTAL_MANAGER.onPortalPassed(entity, pos, (ServerWorld) world, axis);
                    }
                }
            } else if (prevEntityPosAlongAxis < entityPosAlongAxis) {
                if (entityPosAlongAxis > centerPosAlongAxis
                    && prevEntityPosAlongAxis <= centerPosAlongAxis) {
                    double plainHitAfter = (centerPosAlongAxis - entityPosAlongAxis) / movementAlongAxis;
                    Vec3d planeHitPoint = entityPos.add(movement.multiply(plainHitAfter));
                    if (Box.from(new BlockBox(pos)).contains(planeHitPoint)) {
                        PORTAL_MANAGER.onPortalPassed(entity, pos, (ServerWorld) world, axis);
                    }
                }
            }
        }

    }

    public VoxelShape getShape(BlockState state, BlockView world, BlockPos pos) {
        return switch (state.get(AXIS)) {
            case Z -> Z_AABB;
            case Y -> Y_AABB;
            default -> X_AABB;
        };
    }

    @Override
    public VoxelShape getCollisionShape(BlockState state, BlockView world, BlockPos pos, ShapeContext context) {
        if (context instanceof CollisionAwareShapeContext) {
            return getShape(state, world, pos);
        }
        return VoxelShapes.empty();
    }

    @Override
    public BlockState rotate(BlockState state, BlockRotation rotation) {
        return changeRotation(state, rotation);
    }

    public static BlockState changeRotation(BlockState state, BlockRotation rotation) {
        switch (rotation) {
            case COUNTERCLOCKWISE_90, CLOCKWISE_90 -> {
                switch (state.get(AXIS)) {
                    case X -> {
                        return state.with(AXIS, Direction.Axis.Z);
                    }
                    case Z -> {
                        return state.with(AXIS, Direction.Axis.X);
                    }
                }
            }
        }
        return state;
    }

    public BlockState getStateWith(DyeColor color, Direction.Axis axis) {
        return this.getDefaultState().with(AXIS, axis).with(DYE_COLOR, color);

    }

    @Override
    public BlockState getPlacementState(ItemPlacementContext ctx) {
        return this.getDefaultState().with(AXIS, ctx.getPlayerLookDirection().getAxis());
    }

    @Override
    protected void appendProperties(StateManager.Builder<Block, BlockState> builder) {
        builder.add(AXIS).add(DYE_COLOR);
    }

    @Override
    public boolean isTransparent(BlockState state, BlockView world, BlockPos pos) {
        return true;
    }

    @Override
    public boolean canFillWithFluid(@Nullable PlayerEntity player, BlockView world, BlockPos pos, BlockState state, Fluid fluid) {
        if (((World) world).isClient() && fluid == ColourfulPortalsMod.PORTAL_FLUID) {
            return true;
        }
        return fluid == ColourfulPortalsMod.PORTAL_FLUID
               && (world instanceof ServerWorld)
               && PORTAL_MANAGER.canExtend((ServerWorld) world, pos);
    }


    @Override
    public boolean tryFillWithFluid(WorldAccess world, BlockPos pos, BlockState state, FluidState fluidState) {
        if (world.isClient() && fluidState.isOf(ColourfulPortalsMod.PORTAL_FLUID)) {
            return true;
        }
        return fluidState.isOf(ColourfulPortalsMod.PORTAL_FLUID)
               && (world instanceof ServerWorld)
               && PORTAL_MANAGER.extend((ServerWorld) world, pos);
    }
}
