package quimufu.colourful_portals.portal;

import net.minecraft.entity.Entity;
import net.minecraft.registry.RegistryKey;
import net.minecraft.registry.RegistryKeys;
import net.minecraft.server.MinecraftServer;
import net.minecraft.server.world.ServerWorld;
import net.minecraft.util.Identifier;
import net.minecraft.util.TypeFilter;
import net.minecraft.util.math.*;
import net.minecraft.world.World;
import qouteall.imm_ptl.core.api.PortalAPI;
import qouteall.q_misc_util.my_util.DQuaternion;
import quimufu.colourful_portals.general_util.LinkedList;

import java.util.ArrayList;
import java.util.List;

import static quimufu.colourful_portals.ColourfulPortalsMod.LOGGER;
import static quimufu.colourful_portals.ColourfulPortalsMod.MOD_ID;

public class ImmersivePortalsLinkingSystem implements PortalLinkingSystem {

    public static final Identifier IMMERSIVE_PORTALS_LINKING_SYSTEM = Identifier.of(MOD_ID, "immersive_portals_linking_system");

    private final MinecraftServer server;

    public ImmersivePortalsLinkingSystem(MinecraftServer server) {
        this.server = server;
    }

    @Override
    public Identifier getLinkingSystemId() {
        return IMMERSIVE_PORTALS_LINKING_SYSTEM;
    }

    public void assureLinked(PortalRepresentation fromPortalRepresentation, PortalRepresentation linkedToPortalRepresentation) {

        ServerWorld fromPortalWorld = getPortalWorld(fromPortalRepresentation);
        RegistryKey<World> linkedToPortalWorldRegKey = RegistryKey.of(RegistryKeys.WORLD, linkedToPortalRepresentation.dimensionId());

        BlockBox fromPortalBlockBox = fromPortalRepresentation.location();
        Box fromPortalBox = Box.from(fromPortalBlockBox);
        BlockBox linkedToPortalBlockBox = linkedToPortalRepresentation.location();
        Box linkedToPortalBox = Box.from(linkedToPortalBlockBox);

        List<qouteall.imm_ptl.core.portal.Portal> outgoingPortals = new ArrayList<>(getPortalList(fromPortalRepresentation));

        if (outgoingPortals.size() > 2) {
            LOGGER.warn("Found more then 2 portals in {}, cleaning up", fromPortalBox);
            for (int i = 2; i < outgoingPortals.size(); i++) {
                qouteall.imm_ptl.core.portal.Portal outgoingPortal = outgoingPortals.get(i);
                outgoingPortal.kill();
            }
        }
        if (outgoingPortals.isEmpty()) {
            qouteall.imm_ptl.core.portal.Portal portal = qouteall.imm_ptl.core.portal.Portal.ENTITY_TYPE.create(fromPortalWorld);
            if (portal == null) {
                LOGGER.error("could not create PortalRepresentation entity for  {}", fromPortalRepresentation);
                return;
            }

            portal.setOriginPos(fromPortalBox.getCenter());
            portal.setDestinationDimension(linkedToPortalWorldRegKey);
            portal.setDestination(linkedToPortalBox.getCenter());
            Vec3d axisW = Vec3d.of(PortalHelper.getAxisW(fromPortalBlockBox).getVector());
            portal.setOrientationAndSize(
                    axisW, // axisW
                    new Vec3d(0, 1, 0), // axisH
                    2, // width
                    3 // height
            );
            portal.setRotationTransformation(DQuaternion.getRotationBetween(axisW, Vec3d.of(PortalHelper.getAxisW(linkedToPortalBlockBox).getVector())));
            outgoingPortals.add(portal);
            if (!fromPortalWorld.spawnEntity(portal)) {
                LOGGER.error("could not spawn PortalRepresentation entity for  {}", fromPortalRepresentation);
                return;
            }
        }
        if (outgoingPortals.size() == 1) {
            qouteall.imm_ptl.core.portal.Portal portal = PortalAPI.createFlippedPortal(outgoingPortals.getFirst());
            outgoingPortals.add(portal);
            if (!fromPortalWorld.spawnEntity(portal)) {
                LOGGER.error("could not spawn second PortalRepresentation entity for  {}", fromPortalRepresentation);
                return;
            }
        }
        for (qouteall.imm_ptl.core.portal.Portal outgoingPortal : outgoingPortals) {
            outgoingPortal.setDestinationDimension(linkedToPortalWorldRegKey);
            outgoingPortal.setDestination(linkedToPortalBox.getCenter());
            Vec3d axisW = Vec3d.of(PortalHelper.getAxisW(fromPortalBlockBox).getVector());
            outgoingPortal.setRotationTransformation(DQuaternion.getRotationBetween(axisW, Vec3d.of(PortalHelper.getAxisW(linkedToPortalBlockBox).getVector())));
            outgoingPortal.reloadAndSyncToClient();
        }
    }

    @Override
    public void linkPortals(LinkedList<PortalRepresentation> portalRepresentations) {
        for (int i = 0; i < portalRepresentations.size(); i++) {
            PortalRepresentation portalRepresentation = portalRepresentations.get(i);
            PortalRepresentation linkedToPortalRepresentation = i + 1 < portalRepresentations.size() ? portalRepresentations.get(i + 1) : portalRepresentations.getFirst();

            assureLinked(portalRepresentation, linkedToPortalRepresentation);
        }
    }

    @Override
    public void unLinkPortal(PortalRepresentation portalRepresentation) {
        List<qouteall.imm_ptl.core.portal.Portal> portals = getPortalList(portalRepresentation);
        portals.forEach(portalEntity -> PortalAPI.removeGlobalPortal(getPortalWorld(portalRepresentation), portalEntity));
    }

    @Override
    public boolean onPortalPassed(Entity entity, BlockPos pos, ServerWorld world, Direction.Axis a) {
        //no-op, handled by immptl
        return false;
    }

    @Override
    public boolean needsReInit() {
        return false;
    }

    @Override
    public boolean needsMovementCallback() {
        return false;
    }

    @Override
    public boolean movementCallback(Entity entity, BlockPos p, ServerWorld world) {
        return false;
    }

    private List<qouteall.imm_ptl.core.portal.Portal> getPortalList(PortalRepresentation portalRepresentation) {
        Box portalBox = Box.from(portalRepresentation.location());
        return getPortalWorld(portalRepresentation)
                .getEntitiesByType(TypeFilter.instanceOf(qouteall.imm_ptl.core.portal.Portal.class), portalBox, e -> e.isAlive() && contains(portalBox, e.getBoundingBox()));
    }

    private boolean contains(Box outside, Box inside) {
        return outside.contains(inside.getMaxPos()) && outside.contains(inside.getMinPos());
    }

    private ServerWorld getPortalWorld(PortalRepresentation fromPortalRepresentation) {
        ServerWorld serverWorld = PortalHelper.getPortalWorld(server, fromPortalRepresentation);
        if (serverWorld == null) {
            LOGGER.warn("couldn't get portal dimensionId for portal {}. Don't sue me!", fromPortalRepresentation);
            throw new RuntimeException();
        }
        return serverWorld;
    }
}
